@@ -3,8 +3,24 @@
 See header file: geos-x.y.z/capi/geos_c.h
 '''
 
-from ctypes import CFUNCTYPE, POINTER, c_void_p, c_char_p, \
-    c_size_t, c_byte, c_char, c_uint, c_int, c_double, py_object
+from ctypes import (
+    CFUNCTYPE,
+    POINTER,
+    Structure,
+    c_void_p,
+    c_char_p,
+    c_size_t,
+    c_byte,
+    c_uint,
+    c_int,
+    c_double,
+    py_object,
+)
+
+from .errors import UnsupportedGEOSVersionError
+
+
+EXCEPTION_HANDLER_FUNCTYPE = CFUNCTYPE(None, c_char_p, c_void_p)
 
 # Derived pointer types
 c_size_t_p = POINTER(c_size_t)
@@ -14,21 +30,29 @@ class allocated_c_char_p(c_char_p):
     '''char pointer return type'''
     pass
 
-EXCEPTION_HANDLER_FUNCTYPE = CFUNCTYPE(None, c_char_p, c_void_p)
+
+class GEOSGeom_t(Structure):
+    pass
+
+
+c_geom_p = POINTER(GEOSGeom_t)
 
 
 def prototype(lgeos, geos_version):
-    '''Protype functions in geos_c.h for different version of GEOS
+    """Protype functions in geos_c.h for different version of GEOS
 
     Use the GEOS version, not the C API version.
-    '''
+    """
 
-    '''
-    Initialization, cleanup, version
-    '''
+    if not geos_version >= (3, 3, 0):
+        raise UnsupportedGEOSVersionError(
+            "Shapely requires GEOS version 3.3.0 or newer.")
+
+    # Initialization, cleanup, version.
 
     lgeos.initGEOS.restype = None
-    lgeos.initGEOS.argtypes = [EXCEPTION_HANDLER_FUNCTYPE, EXCEPTION_HANDLER_FUNCTYPE]
+    lgeos.initGEOS.argtypes = [
+        EXCEPTION_HANDLER_FUNCTYPE, EXCEPTION_HANDLER_FUNCTYPE]
 
     lgeos.finishGEOS.restype = None
     lgeos.finishGEOS.argtypes = []
@@ -36,10 +60,8 @@ def prototype(lgeos, geos_version):
     lgeos.GEOSversion.restype = c_char_p
     lgeos.GEOSversion.argtypes = []
 
-    '''
-    NOTE - These functions are DEPRECATED.  Please use the new Reader and
-    writer APIS!
-    '''
+    # These functions are DEPRECATED.  Please use the new Reader and
+    # writer APIS!
 
     lgeos.GEOSGeomFromWKT.restype = c_void_p
     lgeos.GEOSGeomFromWKT.argtypes = [c_char_p]
@@ -56,9 +78,7 @@ def prototype(lgeos, geos_version):
     lgeos.GEOSGeomToWKB_buf.restype = allocated_c_char_p
     lgeos.GEOSGeomToWKB_buf.argtypes = [c_void_p, c_size_t_p]
 
-    '''
-    Coordinate sequence
-    '''
+    # Coordinate sequence
 
     lgeos.GEOSCoordSeq_create.restype = c_void_p
     lgeos.GEOSCoordSeq_create.argtypes = [c_uint, c_uint]
@@ -82,7 +102,8 @@ def prototype(lgeos, geos_version):
     lgeos.GEOSCoordSeq_setZ.argtypes = [c_void_p, c_uint, c_double]
 
     lgeos.GEOSCoordSeq_setOrdinate.restype = c_int
-    lgeos.GEOSCoordSeq_setOrdinate.argtypes = [c_void_p, c_uint, c_uint, c_double]
+    lgeos.GEOSCoordSeq_setOrdinate.argtypes = [
+        c_void_p, c_uint, c_uint, c_double]
 
     lgeos.GEOSCoordSeq_getX.restype = c_int
     lgeos.GEOSCoordSeq_getX.argtypes = [c_void_p, c_uint, c_void_p]
@@ -99,72 +120,87 @@ def prototype(lgeos, geos_version):
     lgeos.GEOSCoordSeq_getDimensions.restype = c_int
     lgeos.GEOSCoordSeq_getDimensions.argtypes = [c_void_p, c_void_p]
 
-    '''
-    Linear refeferencing
-    '''
+    # Linear referencing
 
-    if geos_version >= (3, 2, 0):
+    lgeos.GEOSProject.restype = c_double
+    lgeos.GEOSProject.argtypes = [c_void_p, c_void_p]
 
-        lgeos.GEOSProject.restype = c_double
-        lgeos.GEOSProject.argtypes = [c_void_p, c_void_p]
+    lgeos.GEOSInterpolate.restype = c_void_p
+    lgeos.GEOSInterpolate.argtypes = [c_void_p, c_double]
 
-        lgeos.GEOSInterpolate.restype = c_void_p
-        lgeos.GEOSInterpolate.argtypes = [c_void_p, c_double]
+    lgeos.GEOSProjectNormalized.restype = c_double
+    lgeos.GEOSProjectNormalized.argtypes = [c_void_p, c_void_p]
 
-        lgeos.GEOSProjectNormalized.restype = c_double
-        lgeos.GEOSProjectNormalized.argtypes = [c_void_p, c_void_p]
+    lgeos.GEOSInterpolateNormalized.restype = c_void_p
+    lgeos.GEOSInterpolateNormalized.argtypes = [c_void_p, c_double]
 
-        lgeos.GEOSInterpolateNormalized.restype = c_void_p
-        lgeos.GEOSInterpolateNormalized.argtypes = [c_void_p, c_double]
-
-    '''
-    Buffer related
-    '''
+    # Buffer related
 
     lgeos.GEOSBuffer.restype = c_void_p
     lgeos.GEOSBuffer.argtypes = [c_void_p, c_double, c_int]
 
-    if geos_version >= (3, 2, 0):
+    lgeos.GEOSBufferWithStyle.restype = c_void_p
+    lgeos.GEOSBufferWithStyle.argtypes = [
+        c_void_p, c_double, c_int, c_int, c_int, c_double]
 
-        lgeos.GEOSBufferWithStyle.restype = c_void_p
-        lgeos.GEOSBufferWithStyle.argtypes = [c_void_p, c_double, c_int, c_int, c_int, c_double]
+    lgeos.GEOSOffsetCurve.restype = c_void_p
+    lgeos.GEOSOffsetCurve.argtypes = [
+        c_void_p, c_double, c_int, c_int, c_double]
 
-        lgeos.GEOSSingleSidedBuffer.restype = c_void_p
-        lgeos.GEOSSingleSidedBuffer.argtypes = [c_void_p, c_double, c_int, c_int, c_double, c_int]
+    lgeos.GEOSBufferParams_create.restype = c_void_p
+    lgeos.GEOSBufferParams_create.argtypes = None
 
-    '''
-    Geometry constructors
-    '''
+    lgeos.GEOSBufferParams_destroy.restype = None
+    lgeos.GEOSBufferParams_destroy.argtypes = [c_void_p]
 
-    lgeos.GEOSGeom_createPoint.restype = c_void_p
+    lgeos.GEOSBufferParams_setEndCapStyle.restype = c_int
+    lgeos.GEOSBufferParams_setEndCapStyle.argtypes = [c_void_p, c_int]
+
+    lgeos.GEOSBufferParams_setJoinStyle.restype = c_int
+    lgeos.GEOSBufferParams_setJoinStyle.argtypes = [c_void_p, c_int]
+
+    lgeos.GEOSBufferParams_setMitreLimit.restype = c_int
+    lgeos.GEOSBufferParams_setMitreLimit.argtypes = [c_void_p, c_double]
+
+    lgeos.GEOSBufferParams_setQuadrantSegments.restype = c_int
+    lgeos.GEOSBufferParams_setQuadrantSegments.argtypes = [c_void_p, c_int]
+
+    lgeos.GEOSBufferParams_setSingleSided.restype = c_int
+    lgeos.GEOSBufferParams_setSingleSided.argtypes = [c_void_p, c_int]
+
+    lgeos.GEOSBufferWithParams.restype = c_void_p
+    lgeos.GEOSBufferWithParams.argtypes = [c_void_p, c_void_p, c_double]
+
+    # Geometry constructors
+
+    lgeos.GEOSGeom_createPoint.restype = c_geom_p
     lgeos.GEOSGeom_createPoint.argtypes = [c_void_p]
 
-    lgeos.GEOSGeom_createLinearRing.restype = c_void_p
+    lgeos.GEOSGeom_createLinearRing.restype = c_geom_p
     lgeos.GEOSGeom_createLinearRing.argtypes = [c_void_p]
 
-    lgeos.GEOSGeom_createLineString.restype = c_void_p
+    lgeos.GEOSGeom_createLineString.restype = c_geom_p
     lgeos.GEOSGeom_createLineString.argtypes = [c_void_p]
 
-    lgeos.GEOSGeom_createPolygon.restype = c_void_p
-    lgeos.GEOSGeom_createPolygon.argtypes = [c_void_p, c_void_p, c_uint]
+    lgeos.GEOSGeom_createPolygon.restype = c_geom_p
+    lgeos.GEOSGeom_createPolygon.argtypes = [c_geom_p, POINTER(c_geom_p), c_uint]
 
-    lgeos.GEOSGeom_createCollection.restype = c_void_p
-    lgeos.GEOSGeom_createCollection.argtypes = [c_int, c_void_p, c_uint]
+    lgeos.GEOSGeom_createCollection.restype = c_geom_p
+    lgeos.GEOSGeom_createCollection.argtypes = [c_int, POINTER(c_geom_p), c_uint]
 
-    lgeos.GEOSGeom_clone.restype = c_void_p
+    lgeos.GEOSGeom_createEmptyCollection.restype = c_geom_p
+    lgeos.GEOSGeom_createEmptyCollection.argtypes = [c_int]
+
+    lgeos.GEOSGeom_clone.restype = c_geom_p
     lgeos.GEOSGeom_clone.argtypes = [c_void_p]
 
-    '''
-    Memory management
-    '''
+    # Memory management
 
     lgeos.GEOSGeom_destroy.restype = None
     lgeos.GEOSGeom_destroy.argtypes = [c_void_p]
 
-    '''
-    Topology operations
-    Return NULL on exception
-    '''
+    # Topology operations
+    # Return NULL on exception
 
     lgeos.GEOSEnvelope.restype = c_void_p
     lgeos.GEOSEnvelope.argtypes = [c_void_p]
@@ -187,14 +223,12 @@ def prototype(lgeos, geos_version):
     lgeos.GEOSUnion.restype = c_void_p
     lgeos.GEOSUnion.argtypes = [c_void_p, c_void_p]
 
-    if geos_version >= (3, 3, 0):
-        lgeos.GEOSUnaryUnion.restype = c_void_p
-        lgeos.GEOSUnaryUnion.argtypes = [c_void_p]
+    lgeos.GEOSUnaryUnion.restype = c_void_p
+    lgeos.GEOSUnaryUnion.argtypes = [c_void_p]
 
-    if geos_version >= (3, 1, 0):
-        '''deprecated in 3.3.0: use GEOSUnaryUnion instead'''
-        lgeos.GEOSUnionCascaded.restype = c_void_p
-        lgeos.GEOSUnionCascaded.argtypes = [c_void_p]
+    # deprecated in 3.3.0: use GEOSUnaryUnion instead
+    lgeos.GEOSUnionCascaded.restype = c_void_p
+    lgeos.GEOSUnionCascaded.argtypes = [c_void_p]
 
     lgeos.GEOSPointOnSurface.restype = c_void_p
     lgeos.GEOSPointOnSurface.argtypes = [c_void_p]
@@ -202,17 +236,28 @@ def prototype(lgeos, geos_version):
     lgeos.GEOSGetCentroid.restype = c_void_p
     lgeos.GEOSGetCentroid.argtypes = [c_void_p]
 
+    if geos_version >= (3, 5, 0):
+        lgeos.GEOSClipByRect.restype = c_void_p
+        lgeos.GEOSClipByRect.argtypes = [c_void_p, c_double, c_double, c_double, c_double]
+
     lgeos.GEOSPolygonize.restype = c_void_p
     lgeos.GEOSPolygonize.argtypes = [c_void_p, c_uint]
 
-    if geos_version >= (3, 3, 0):
-        lgeos.GEOSPolygonize_full.restype = c_void_p
-        lgeos.GEOSPolygonize_full.argtypes = [c_void_p, c_void_p, c_void_p, c_void_p]
+    lgeos.GEOSPolygonize_full.restype = c_geom_p
+    lgeos.GEOSPolygonize_full.argtypes = [
+        c_geom_p,
+        POINTER(c_geom_p),
+        POINTER(c_geom_p),
+        POINTER(c_geom_p),
+    ]
 
     if geos_version >= (3, 4, 0):
         lgeos.GEOSDelaunayTriangulation.restype = c_void_p
         lgeos.GEOSDelaunayTriangulation.argtypes = [c_void_p, c_double, c_int]
 
+    if geos_version >= (3, 5, 0):
+        lgeos.GEOSVoronoiDiagram.restype = c_void_p
+        lgeos.GEOSVoronoiDiagram.argtypes = [c_void_p, c_void_p, c_double, c_int]
 
     lgeos.GEOSLineMerge.restype = c_void_p
     lgeos.GEOSLineMerge.argtypes = [c_void_p]
@@ -223,10 +268,8 @@ def prototype(lgeos, geos_version):
     lgeos.GEOSTopologyPreserveSimplify.restype = c_void_p
     lgeos.GEOSTopologyPreserveSimplify.argtypes = [c_void_p, c_double]
 
-    '''
-    Binary predicates
-    Return 2 on exception, 1 on true, 0 on false
-    '''
+    # Binary predicates
+    # Return 2 on exception, 1 on true, 0 on false
 
     lgeos.GEOSDisjoint.restype = c_byte
     lgeos.GEOSDisjoint.argtypes = [c_void_p, c_void_p]
@@ -249,16 +292,20 @@ def prototype(lgeos, geos_version):
     lgeos.GEOSOverlaps.restype = c_byte
     lgeos.GEOSOverlaps.argtypes = [c_void_p, c_void_p]
 
+    lgeos.GEOSCovers.restype = c_byte
+    lgeos.GEOSCovers.argtypes = [c_void_p, c_void_p]
+
+    lgeos.GEOSCoveredBy.restype = c_byte
+    lgeos.GEOSCoveredBy.argtypes = [c_void_p, c_void_p]
+
     lgeos.GEOSEquals.restype = c_byte
     lgeos.GEOSEquals.argtypes = [c_void_p, c_void_p]
 
     lgeos.GEOSEqualsExact.restype = c_byte
     lgeos.GEOSEqualsExact.argtypes = [c_void_p, c_void_p, c_double]
 
-    '''
-    Unary predicate
-    Return 2 on exception, 1 on true, 0 on false
-    '''
+    # Unary predicate
+    # Return 2 on exception, 1 on true, 0 on false
 
     lgeos.GEOSisEmpty.restype = c_byte
     lgeos.GEOSisEmpty.argtypes = [c_void_p]
@@ -266,9 +313,8 @@ def prototype(lgeos, geos_version):
     lgeos.GEOSisValid.restype = c_byte
     lgeos.GEOSisValid.argtypes = [c_void_p]
 
-    if geos_version >= (3, 1, 0):
-        lgeos.GEOSisValidReason.restype = allocated_c_char_p
-        lgeos.GEOSisValidReason.argtypes = [c_void_p]
+    lgeos.GEOSisValidReason.restype = allocated_c_char_p
+    lgeos.GEOSisValidReason.argtypes = [c_void_p]
 
     lgeos.GEOSisSimple.restype = c_byte
     lgeos.GEOSisSimple.argtypes = [c_void_p]
@@ -276,51 +322,60 @@ def prototype(lgeos, geos_version):
     lgeos.GEOSisRing.restype = c_byte
     lgeos.GEOSisRing.argtypes = [c_void_p]
 
-    if geos_version >= (3, 3, 0):
-        lgeos.GEOSisClosed.restype = c_byte
-        lgeos.GEOSisClosed.argtypes = [c_void_p]
+    lgeos.GEOSisClosed.restype = c_byte
+    lgeos.GEOSisClosed.argtypes = [c_void_p]
 
     lgeos.GEOSHasZ.restype = c_byte
     lgeos.GEOSHasZ.argtypes = [c_void_p]
 
-    '''
-    Dimensionally Extended 9 Intersection Model related
-    '''
-
-    lgeos.GEOSRelatePattern.restype = c_char
-    lgeos.GEOSRelatePattern.argtypes = [c_void_p, c_void_p, c_char_p]
+    # Dimensionally Extended 9 Intersection Model related
 
     lgeos.GEOSRelate.restype = allocated_c_char_p
     lgeos.GEOSRelate.argtypes = [c_void_p, c_void_p]
 
-    '''
-    Prepared Geometry Binary predicates
-    Return 2 on exception, 1 on true, 0 on false
-    '''
+    lgeos.GEOSRelatePattern.restype = c_byte
+    lgeos.GEOSRelatePattern.argtypes = [c_void_p, c_void_p, c_char_p]
 
-    if geos_version >= (3, 1, 0):
+    lgeos.GEOSRelatePatternMatch.restype = c_byte
+    lgeos.GEOSRelatePatternMatch.argtypes = [c_char_p, c_char_p]
 
-        lgeos.GEOSPrepare.restype = c_void_p
-        lgeos.GEOSPrepare.argtypes = [c_void_p]
+    # Prepared Geometry Binary predicates
+    # Return 2 on exception, 1 on true, 0 on false
 
-        lgeos.GEOSPreparedGeom_destroy.restype = None
-        lgeos.GEOSPreparedGeom_destroy.argtypes = [c_void_p]
+    lgeos.GEOSPrepare.restype = c_void_p
+    lgeos.GEOSPrepare.argtypes = [c_void_p]
 
-        lgeos.GEOSPreparedContains.restype = c_int
-        lgeos.GEOSPreparedContains.argtypes = [c_void_p, c_void_p]
+    lgeos.GEOSPreparedGeom_destroy.restype = None
+    lgeos.GEOSPreparedGeom_destroy.argtypes = [c_void_p]
 
-        lgeos.GEOSPreparedContainsProperly.restype = c_int
-        lgeos.GEOSPreparedContainsProperly.argtypes = [c_void_p, c_void_p]
+    lgeos.GEOSPreparedDisjoint.restype = c_byte
+    lgeos.GEOSPreparedDisjoint.argtypes = [c_void_p, c_void_p]
 
-        lgeos.GEOSPreparedCovers.restype = c_int
-        lgeos.GEOSPreparedCovers.argtypes = [c_void_p, c_void_p]
+    lgeos.GEOSPreparedTouches.restype = c_byte
+    lgeos.GEOSPreparedTouches.argtypes = [c_void_p, c_void_p]
 
-        lgeos.GEOSPreparedIntersects.restype = c_int
-        lgeos.GEOSPreparedIntersects.argtypes = [c_void_p, c_void_p]
+    lgeos.GEOSPreparedIntersects.restype = c_byte
+    lgeos.GEOSPreparedIntersects.argtypes = [c_void_p, c_void_p]
 
-    '''
-    Geometry info
-    '''
+    lgeos.GEOSPreparedCrosses.restype = c_byte
+    lgeos.GEOSPreparedCrosses.argtypes = [c_void_p, c_void_p]
+
+    lgeos.GEOSPreparedWithin.restype = c_byte
+    lgeos.GEOSPreparedWithin.argtypes = [c_void_p, c_void_p]
+
+    lgeos.GEOSPreparedContains.restype = c_byte
+    lgeos.GEOSPreparedContains.argtypes = [c_void_p, c_void_p]
+
+    lgeos.GEOSPreparedContainsProperly.restype = c_byte
+    lgeos.GEOSPreparedContainsProperly.argtypes = [c_void_p, c_void_p]
+
+    lgeos.GEOSPreparedOverlaps.restype = c_byte
+    lgeos.GEOSPreparedOverlaps.argtypes = [c_void_p, c_void_p]
+
+    lgeos.GEOSPreparedCovers.restype = c_byte
+    lgeos.GEOSPreparedCovers.argtypes = [c_void_p, c_void_p]
+
+    # Geometry info
 
     lgeos.GEOSGeomType.restype = c_char_p
     lgeos.GEOSGeomType.argtypes = [c_void_p]
@@ -358,24 +413,25 @@ def prototype(lgeos, geos_version):
     lgeos.GEOSGeom_getDimensions.restype = c_int
     lgeos.GEOSGeom_getDimensions.argtypes = [c_void_p]
 
-    '''
-    Misc functions
-    '''
+    lgeos.GEOSNormalize.restype = c_int
+    lgeos.GEOSNormalize.argtypes = [c_void_p]
 
-    lgeos.GEOSArea.restype = c_double
-    lgeos.GEOSArea.argtypes = [c_void_p, c_void_p]
+    # Misc functions
+
+    lgeos.GEOSArea.restype = c_int
+    lgeos.GEOSArea.argtypes = [c_geom_p, POINTER(c_double)]
 
     lgeos.GEOSLength.restype = c_int
-    lgeos.GEOSLength.argtypes = [c_void_p, c_void_p]
+    lgeos.GEOSLength.argtypes = [c_geom_p, POINTER(c_double)]
 
     lgeos.GEOSDistance.restype = c_int
-    lgeos.GEOSDistance.argtypes = [c_void_p, c_void_p, c_void_p]
+    lgeos.GEOSDistance.argtypes = [c_geom_p, c_geom_p, POINTER(c_double)]
 
-    '''
-    Reader and Writer APIs
-    '''
+    lgeos.GEOSHausdorffDistance.restype = c_int
+    lgeos.GEOSHausdorffDistance.argtypes = [c_geom_p, c_geom_p, POINTER(c_double)]
+
+    # Reader and Writer APIs
 
-    '''WKT Reader'''
     lgeos.GEOSWKTReader_create.restype = c_void_p
     lgeos.GEOSWKTReader_create.argtypes = []
 
@@ -385,7 +441,6 @@ def prototype(lgeos, geos_version):
     lgeos.GEOSWKTReader_read.restype = c_void_p
     lgeos.GEOSWKTReader_read.argtypes = [c_void_p, c_char_p]
 
-    '''WKT Writer'''
     lgeos.GEOSWKTWriter_create.restype = c_void_p
     lgeos.GEOSWKTWriter_create.argtypes = []
 
@@ -395,24 +450,21 @@ def prototype(lgeos, geos_version):
     lgeos.GEOSWKTWriter_write.restype = allocated_c_char_p
     lgeos.GEOSWKTWriter_write.argtypes = [c_void_p, c_void_p]
 
-    if geos_version >= (3, 3, 0):
+    lgeos.GEOSWKTWriter_setTrim.restype = None
+    lgeos.GEOSWKTWriter_setTrim.argtypes = [c_void_p, c_int]
 
-        lgeos.GEOSWKTWriter_setTrim.restype = None
-        lgeos.GEOSWKTWriter_setTrim.argtypes = [c_void_p, c_int]
+    lgeos.GEOSWKTWriter_setRoundingPrecision.restype = None
+    lgeos.GEOSWKTWriter_setRoundingPrecision.argtypes = [c_void_p, c_int]
 
-        lgeos.GEOSWKTWriter_setRoundingPrecision.restype = None
-        lgeos.GEOSWKTWriter_setRoundingPrecision.argtypes = [c_void_p, c_int]
+    lgeos.GEOSWKTWriter_setOutputDimension.restype = None
+    lgeos.GEOSWKTWriter_setOutputDimension.argtypes = [c_void_p, c_int]
 
-        lgeos.GEOSWKTWriter_setOutputDimension.restype = None
-        lgeos.GEOSWKTWriter_setOutputDimension.argtypes = [c_void_p, c_int]
+    lgeos.GEOSWKTWriter_getOutputDimension.restype = c_int
+    lgeos.GEOSWKTWriter_getOutputDimension.argtypes = [c_void_p]
 
-        lgeos.GEOSWKTWriter_getOutputDimension.restype = c_int
-        lgeos.GEOSWKTWriter_getOutputDimension.argtypes = [c_void_p]
+    lgeos.GEOSWKTWriter_setOld3D.restype = None
+    lgeos.GEOSWKTWriter_setOld3D.argtypes = [c_void_p, c_int]
 
-        lgeos.GEOSWKTWriter_setOld3D.restype = None
-        lgeos.GEOSWKTWriter_setOld3D.argtypes = [c_void_p, c_int]
-
-    '''WKB Reader'''
     lgeos.GEOSWKBReader_create.restype = c_void_p
     lgeos.GEOSWKBReader_create.argtypes = []
 
@@ -425,7 +477,6 @@ def prototype(lgeos, geos_version):
     lgeos.GEOSWKBReader_readHEX.restype = c_void_p
     lgeos.GEOSWKBReader_readHEX.argtypes = [c_void_p, c_char_p, c_size_t]
 
-    '''WKB Writer'''
     lgeos.GEOSWKBWriter_create.restype = c_void_p
     lgeos.GEOSWKBWriter_create.argtypes = []
 
@@ -456,19 +507,14 @@ def prototype(lgeos, geos_version):
     lgeos.GEOSWKBWriter_setIncludeSRID.restype = None
     lgeos.GEOSWKBWriter_setIncludeSRID.argtypes = [c_void_p, c_int]
 
-    if geos_version >= (3, 1, 1):
+    lgeos.GEOSFree.restype = None
+    lgeos.GEOSFree.argtypes = [c_void_p]
 
-        '''
-        Free buffers returned by stuff like GEOSWKBWriter_write(),
-        GEOSWKBWriter_writeHEX() and GEOSWKTWriter_write()
-        '''
+    lgeos.GEOSSnap.restype = c_void_p
+    lgeos.GEOSSnap.argtypes = [c_void_p, c_void_p, c_double]
 
-        lgeos.GEOSFree.restype = None
-        lgeos.GEOSFree.argtypes = [c_void_p]
-
-    if geos_version >= (3, 3, 0):
-        lgeos.GEOSSnap.restype = c_void_p
-        lgeos.GEOSSnap.argtypes = [c_void_p, c_void_p, c_double]
+    lgeos.GEOSSharedPaths.restype = c_void_p
+    lgeos.GEOSSharedPaths.argtypes = [c_void_p, c_void_p]
 
     if geos_version >= (3, 4, 0):
         lgeos.GEOSNearestPoints.restype = c_void_p
@@ -477,7 +523,8 @@ def prototype(lgeos, geos_version):
     if geos_version >= (3, 4, 2):
         lgeos.GEOSQueryCallback = CFUNCTYPE(None, c_void_p, c_void_p)
 
-        lgeos.GEOSSTRtree_query.argtypes = [c_void_p, c_void_p, lgeos.GEOSQueryCallback, py_object]
+        lgeos.GEOSSTRtree_query.argtypes = [
+            c_void_p, c_void_p, lgeos.GEOSQueryCallback, py_object]
         lgeos.GEOSSTRtree_query.restype = None
 
         lgeos.GEOSSTRtree_create.argtypes = [c_int]
@@ -492,3 +539,17 @@ def prototype(lgeos, geos_version):
         lgeos.GEOSSTRtree_destroy.argtypes = [c_void_p]
         lgeos.GEOSSTRtree_destroy.restype = None
 
+    if geos_version >= (3, 6, 0):
+        lgeos.GEOSDistanceCallback = CFUNCTYPE(c_int, c_void_p, c_void_p, c_void_p, c_void_p)
+
+        lgeos.GEOSSTRtree_nearest_generic.argtypes = [
+            c_void_p, py_object, c_void_p, lgeos.GEOSDistanceCallback, py_object]
+        lgeos.GEOSSTRtree_nearest_generic.restype = c_void_p
+
+        lgeos.GEOSMinimumClearance.argtypes = [c_geom_p, POINTER(c_double)]
+        lgeos.GEOSMinimumClearance.restype = c_int
+
+    if geos_version >= (3, 8, 0):
+        lgeos.GEOSMakeValid.restype = c_void_p
+        lgeos.GEOSMakeValid.argtypes = [c_void_p]
+
