@@ -1,28 +1,37 @@
 """Base geometry class and utilities
+
+Note: a third, z, coordinate value may be used when constructing
+geometry objects, but has no effect on geometric analysis. All
+operations are performed in the x-y plane. Thus, geometries with
+different z values may intersect or be equal.
 """
 
+from binascii import a2b_hex
+from ctypes import cast, pointer, c_size_t, c_char_p, c_void_p
+from itertools import islice
+import logging
+import math
 import sys
 from warnings import warn
-from binascii import a2b_hex
-from ctypes import pointer, c_size_t, c_char_p, c_void_p
+from functools import wraps
+import warnings
 
+from shapely.affinity import affine_transform
 from shapely.coords import CoordinateSequence
-from shapely.ftools import wraps
-from shapely.geos import lgeos, ReadingError
+from shapely.ctypes_declarations import c_geom_p
+from shapely.errors import GeometryTypeError, WKBReadingError, WKTReadingError
+from shapely.errors import ShapelyDeprecationWarning
 from shapely.geos import WKBWriter, WKTWriter
+from shapely.geos import lgeos
 from shapely.impl import DefaultImplementation, delegated
 
-if sys.version_info[0] < 3:
-    range = xrange
-    integer_types = (int, long)
-else:
-    integer_types = (int,)
+log = logging.getLogger(__name__)
 
 try:
     import numpy as np
-    integer_types = integer_types + (np.integer,)
+    integer_types = (int, np.integer)
 except ImportError:
-    pass
+    integer_types = (int,)
 
 
 GEOMETRY_TYPES = [
@@ -48,9 +57,9 @@ def dump_coords(geom):
         return geom.exterior.coords[:] + [i.coords[:] for i in geom.interiors]
     elif geom.type.startswith('Multi') or geom.type == 'GeometryCollection':
         # Recursive call
-        return [dump_coords(part) for part in geom]
+        return [dump_coords(part) for part in geom.geoms]
     else:
-        raise ValueError('Unhandled geometry type: ' + repr(geom.type))
+        raise GeometryTypeError('Unhandled geometry type: ' + repr(geom.type))
 
 
 def geometry_type_name(g):
@@ -73,70 +82,37 @@ def geom_factory(g, parent=None):
         [geom_type],
         )
     ob.__class__ = getattr(mod, geom_type)
-    ob._geom = g
+    ob._set_geom(g)
     ob.__p__ = parent
     if lgeos.methods['has_z'](g):
         ob._ndim = 3
     else:
         ob._ndim = 2
+    ob._is_empty = False
     return ob
 
 
-def geom_from_wkt(data):
-    warn("`geom_from_wkt` is deprecated. Use `geos.wkt_reader.read(data)`.",
-         DeprecationWarning)
-    if sys.version_info[0] >= 3:
-        data = data.encode('ascii')
-    geom = lgeos.GEOSGeomFromWKT(c_char_p(data))
-    if not geom:
-        raise ReadingError(
-            "Could not create geometry because of errors while reading input.")
-    return geom_factory(geom)
-
-
-def geom_to_wkt(ob):
-    warn("`geom_to_wkt` is deprecated. Use `geos.wkt_writer.write(ob)`.",
-         DeprecationWarning)
-    if ob is None or ob._geom is None:
-        raise ValueError("Null geometry supports no operations")
-    return lgeos.GEOSGeomToWKT(ob._geom)
-
-
 def deserialize_wkb(data):
     geom = lgeos.GEOSGeomFromWKB_buf(c_char_p(data), c_size_t(len(data)))
     if not geom:
-        raise ReadingError(
+        raise WKBReadingError(
             "Could not create geometry because of errors while reading input.")
     return geom
 
 
-def geom_from_wkb(data):
-    warn("`geom_from_wkb` is deprecated. Use `geos.wkb_reader.read(data)`.",
-         DeprecationWarning)
-    return geom_factory(deserialize_wkb(data))
-
-
-def geom_to_wkb(ob):
-    warn("`geom_to_wkb` is deprecated. Use `geos.wkb_writer.write(ob)`.",
-         DeprecationWarning)
-    if ob is None or ob._geom is None:
-        raise ValueError("Null geometry supports no operations")
-    size = c_size_t()
-    return lgeos.GEOSGeomToWKB_buf(c_void_p(ob._geom), pointer(size))
-
-
 def geos_geom_from_py(ob, create_func=None):
     """Helper function for geos_*_from_py functions in each geom type.
 
-    If a create_func is specified the coodinate sequence is cloned and a new
+    If a create_func is specified the coordinate sequence is cloned and a new
     geometry is created with it, otherwise the geometry is cloned directly.
     This behaviour is useful for converting between LineString and LinearRing
     objects.
     """
+    geom_ptr = cast(ob._geom, c_geom_p)
     if create_func is None:
-        geom = lgeos.GEOSGeom_clone(ob._geom)
+        geom = lgeos.GEOSGeom_clone(geom_ptr)
     else:
-        cs = lgeos.GEOSGeom_getCoordSeq(ob._geom)
+        cs = lgeos.GEOSGeom_getCoordSeq(geom_ptr)
         cs = lgeos.GEOSCoordSeq_clone(cs)
         geom = create_func(cs)
 
@@ -144,6 +120,7 @@ def geos_geom_from_py(ob, create_func=None):
 
     return geom, N
 
+
 def exceptNull(func):
     """Decorator which helps avoid GEOS operations on null pointers."""
     @wraps(func)
@@ -154,13 +131,13 @@ def exceptNull(func):
     return wrapper
 
 
-class CAP_STYLE(object):
+class CAP_STYLE:
     round = 1
     flat = 2
     square = 3
 
 
-class JOIN_STYLE(object):
+class JOIN_STYLE:
     round = 1
     mitre = 2
     bevel = 3
@@ -168,7 +145,7 @@ class JOIN_STYLE(object):
 EMPTY = deserialize_wkb(a2b_hex(b'010700000000000000'))
 
 
-class BaseGeometry(object):
+class BaseGeometry:
     """
     Provides GEOS spatial predicates and topological operations.
 
@@ -207,18 +184,31 @@ class BaseGeometry(object):
     _lgeos = lgeos
 
     def empty(self, val=EMPTY):
-        # TODO: defer cleanup to the implementation. We shouldn't be
-        # explicitly calling a lgeos method here.
-        if not self._is_empty and not self._other_owned and self.__geom__:
+        warn(
+            "The 'empty()' method is deprecated and will be removed in "
+            "Shapely 2.0",
+            ShapelyDeprecationWarning, stacklevel=2)
+        self._empty(val=val)
+
+    def _empty(self, val=EMPTY):
+        if not self._other_owned and self.__geom__ and self.__geom__ != EMPTY:
             try:
                 self._lgeos.GEOSGeom_destroy(self.__geom__)
-            except AttributeError:
-                pass  # _lgeos might be empty on shutdown
+            except (AttributeError, TypeError):
+                # _lgeos might be empty on shutdown
+                log.exception("Failed to delete GEOS geom")
+
         self._is_empty = True
         self.__geom__ = val
 
+    def __bool__(self):
+        return self.is_empty is False
+
+    def __nonzero__(self):
+        return self.__bool__()
+
     def __del__(self):
-        self.empty(val=None)
+        self._empty(val=None)
         self.__p__ = None
 
     def __str__(self):
@@ -229,8 +219,9 @@ class BaseGeometry(object):
         return (self.__class__, (), self.wkb)
 
     def __setstate__(self, state):
-        self.empty()
+        self._empty()
         self.__geom__ = deserialize_wkb(state)
+        self._is_empty = False
         if lgeos.methods['has_z'](self.__geom__):
             self._ndim = 3
         else:
@@ -242,10 +233,35 @@ class BaseGeometry(object):
 
     @_geom.setter
     def _geom(self, val):
-        self.empty()
+        warn(
+            "Setting the '_geom' attribute directly is deprecated, and will "
+            "no longer be possible in Shapely 2.0.",
+            ShapelyDeprecationWarning, stacklevel=2)
+        self._set_geom(val)
+
+    def _set_geom(self, val):
+        self._empty()
         self._is_empty = val in [EMPTY, None]
         self.__geom__ = val
 
+    def __setattr__(self, name, value):
+        # first try regular attribute access via __getattribute__, so that
+        # our own (existing) attributes don't raise a warning
+        try:
+            object.__getattribute__(self, name)
+            super().__setattr__(name, value)
+            return
+        except AttributeError:
+            pass
+
+        # if custom attribute, raise the warning
+        warn(
+            "Setting custom attributes on geometry objects is deprecated, "
+            "and will raise an AttributeError in Shapely 2.0",
+            ShapelyDeprecationWarning, stacklevel=2
+        )
+        super().__setattr__(name, value)
+
     # Operators
     # ---------
 
@@ -263,25 +279,33 @@ class BaseGeometry(object):
 
     def __eq__(self, other):
         return (
-            isinstance(other, self.__class__) and
+            type(other) == type(self) and
             tuple(self.coords) == tuple(other.coords)
         )
 
     def __ne__(self, other):
         return not self.__eq__(other)
 
-    __hash__ = object.__hash__
+    __hash__ = None
 
     # Array and ctypes interfaces
     # ---------------------------
 
     @property
-    def ctypes(self):
-        """Return ctypes buffer"""
+    def _ctypes(self):
         raise NotImplementedError
 
     @property
-    def array_interface_base(self):
+    def ctypes(self):
+        """Return ctypes buffer"""
+        warn(
+            "Accessing the 'ctypes' attribute is deprecated,"
+            " and will not be possible any more in Shapely 2.0",
+            ShapelyDeprecationWarning, stacklevel=2)
+        return self._ctypes
+
+    @property
+    def _array_interface_base(self):
         if sys.byteorder == 'little':
             typestr = '<f8'
         elif sys.byteorder == 'big':
@@ -292,22 +316,28 @@ class BaseGeometry(object):
         return {
             'version': 3,
             'typestr': typestr,
-            'data': self.ctypes,
+            'data': self._ctypes,
             }
 
+    @property
+    def array_interface_base(self):
+        warn(
+            "The 'array_interface_base' property is deprecated and will be "
+            "removed in Shapely 2.0.",
+            ShapelyDeprecationWarning, stacklevel=2)
+        return self._array_interface_base
+
     @property
     def __array_interface__(self):
         """Provide the Numpy array protocol."""
-        raise NotImplementedError
+        raise AttributeError("Not implemented")
 
     # Coordinate access
     # -----------------
 
     def _get_coords(self):
         """Access to geometry's coordinates (CoordinateSequence)"""
-        if self.is_empty:
-            return []
-        return CoordinateSequence(self)
+        raise NotImplementedError("set_coords must be provided by derived classes")
 
     def _set_coords(self, ob):
         raise NotImplementedError(
@@ -337,20 +367,10 @@ class BaseGeometry(object):
     def type(self):
         return self.geometryType()
 
-    def to_wkb(self):
-        warn("`to_wkb` is deprecated. Use the `wkb` property.",
-             DeprecationWarning)
-        return geom_to_wkb(self)
-
-    def to_wkt(self):
-        warn("`to_wkt` is deprecated. Use the `wkt` property.",
-             DeprecationWarning)
-        return geom_to_wkt(self)
-
     @property
-    def wkt(self, **kw):
+    def wkt(self):
         """WKT representation of the geometry"""
-        return WKTWriter(lgeos, **kw).write(self)
+        return WKTWriter(lgeos).write(self)
 
     @property
     def wkb(self):
@@ -395,8 +415,8 @@ class BaseGeometry(object):
                 scale_factor = max([dx, dy]) / max([width, height])
             except ZeroDivisionError:
                 scale_factor = 1.
-            view_box = "{0} {1} {2} {3}".format(xmin, ymin, dx, dy)
-            transform = "matrix(1,0,0,-1,0,{0})".format(ymax + ymin)
+            view_box = "{} {} {} {}".format(xmin, ymin, dx, dy)
+            transform = "matrix(1,0,0,-1,0,{})".format(ymax + ymin)
             return svg_top + (
                 'width="{1}" height="{2}" viewBox="{0}" '
                 'preserveAspectRatio="xMinYMin meet">'
@@ -421,11 +441,20 @@ class BaseGeometry(object):
         """Unitless distance to other geometry (float)"""
         return self.impl['distance'](self, other)
 
+    def hausdorff_distance(self, other):
+        """Unitless hausdorff distance to other geometry (float)"""
+        return self.impl['hausdorff_distance'](self, other)
+
     @property
     def length(self):
         """Unitless length of the geometry (float)"""
         return self.impl['length'](self)
 
+    @property
+    def minimum_clearance(self):
+        """Unitless distance by which a node could be moved to produce an invalid geometry (float)"""
+        return self.impl['minimum_clearance'](self)
+
     # Topological properties
     # ----------------------
 
@@ -472,46 +501,117 @@ class BaseGeometry(object):
         """A figure that envelopes the geometry"""
         return geom_factory(self.impl['envelope'](self))
 
+    @property
+    def minimum_rotated_rectangle(self):
+        """Returns the general minimum bounding rectangle of
+        the geometry. Can possibly be rotated. If the convex hull
+        of the object is a degenerate (line or point) this same degenerate
+        is returned.
+        """
+        # first compute the convex hull
+        hull = self.convex_hull
+        try:
+            coords = hull.exterior.coords
+        except AttributeError:  # may be a Point or a LineString
+            return hull
+        # generate the edge vectors between the convex hull's coords
+        edges = ((pt2[0] - pt1[0], pt2[1] - pt1[1]) for pt1, pt2 in zip(
+            coords, islice(coords, 1, None)))
+
+        def _transformed_rects():
+            for dx, dy in edges:
+                # compute the normalized direction vector of the edge
+                # vector.
+                length = math.sqrt(dx ** 2 + dy ** 2)
+                ux, uy = dx / length, dy / length
+                # compute the normalized perpendicular vector
+                vx, vy = -uy, ux
+                # transform hull from the original coordinate system to
+                # the coordinate system defined by the edge and compute
+                # the axes-parallel bounding rectangle.
+                transf_rect = affine_transform(
+                    hull, (ux, uy, vx, vy, 0, 0)).envelope
+                # yield the transformed rectangle and a matrix to
+                # transform it back to the original coordinate system.
+                yield (transf_rect, (ux, vx, uy, vy, 0, 0))
+
+        # check for the minimum area rectangle and return it
+        transf_rect, inv_matrix = min(
+            _transformed_rects(), key=lambda r: r[0].area)
+        return affine_transform(transf_rect, inv_matrix)
+
     def buffer(self, distance, resolution=16, quadsegs=None,
                cap_style=CAP_STYLE.round, join_style=JOIN_STYLE.round,
-               mitre_limit=5.0):
-        """Returns a geometry with an envelope at a distance from the object's
-        envelope
+               mitre_limit=5.0, single_sided=False):
+        """Get a geometry that represents all points within a distance
+        of this geometry.
 
-        A negative distance has a "shrink" effect. A zero distance may be used
-        to "tidy" a polygon. The resolution of the buffer around each vertex of
-        the object increases by increasing the resolution keyword parameter
-        or second positional parameter. Note: the use of a `quadsegs` parameter
-        is deprecated and will be gone from the next major release.
+        A positive distance produces a dilation, a negative distance an
+        erosion. A very small or zero distance may sometimes be used to
+        "tidy" a polygon.
 
-        The styles of caps are: CAP_STYLE.round (1), CAP_STYLE.flat (2), and
-        CAP_STYLE.square (3).
+        Parameters
+        ----------
+        distance : float
+            The distance to buffer around the object.
+        resolution : int, optional
+            The resolution of the buffer around each vertex of the
+            object.
+        quadsegs : int, optional
+            Sets the number of line segments used to approximate an
+            angle fillet.  Note: the use of a `quadsegs` parameter is
+            deprecated and will be gone from the next major release.
+        cap_style : int, optional
+            The styles of caps are: CAP_STYLE.round (1), CAP_STYLE.flat
+            (2), and CAP_STYLE.square (3).
+        join_style : int, optional
+            The styles of joins between offset segments are:
+            JOIN_STYLE.round (1), JOIN_STYLE.mitre (2), and
+            JOIN_STYLE.bevel (3).
+        mitre_limit : float, optional
+            The mitre limit ratio is used for very sharp corners. The
+            mitre ratio is the ratio of the distance from the corner to
+            the end of the mitred offset corner. When two line segments
+            meet at a sharp angle, a miter join will extend the original
+            geometry. To prevent unreasonable geometry, the mitre limit
+            allows controlling the maximum length of the join corner.
+            Corners with a ratio which exceed the limit will be beveled.
+        single_side : bool, optional
+            The side used is determined by the sign of the buffer
+            distance:
 
-        The styles of joins between offset segments are: JOIN_STYLE.round (1),
-        JOIN_STYLE.mitre (2), and JOIN_STYLE.bevel (3).
+                a positive distance indicates the left-hand side
+                a negative distance indicates the right-hand side
 
-        The mitre limit ratio is used for very sharp corners. The mitre ratio
-        is the ratio of the distance from the corner to the end of the mitred
-        offset corner. When two line segments meet at a sharp angle, a miter
-        join will extend the original geometry. To prevent unreasonable
-        geometry, the mitre limit allows controlling the maximum length of the
-        join corner. Corners with a ratio which exceed the limit will be
-        beveled.
+            The single-sided buffer of point geometries is the same as
+            the regular buffer.  The End Cap Style for single-sided
+            buffers is always ignored, and forced to the equivalent of
+            CAP_FLAT.
 
-        Example:
+        Returns
+        -------
+        Geometry
+
+        Notes
+        -----
+        The return value is a strictly two-dimensional geometry. All
+        Z coordinates of the original geometry will be ignored.
+
+        Examples
+        --------
+        >>> from shapely.wkt import loads
+        >>> g = loads('POINT (0.0 0.0)')
+        >>> g.buffer(1.0).area        # 16-gon approx of a unit radius circle
+        3.1365484905459...
+        >>> g.buffer(1.0, 128).area   # 128-gon approximation
+        3.141513801144...
+        >>> round(g.buffer(1.0, 3).area, 10)  # triangle approximation
+        3.0
+        >>> list(g.buffer(1.0, cap_style=CAP_STYLE.square).exterior.coords)
+        [(1.0, 1.0), (1.0, -1.0), (-1.0, -1.0), (-1.0, 1.0), (1.0, 1.0)]
+        >>> g.buffer(1.0, cap_style=CAP_STYLE.square).area
+        4.0
 
-          >>> from shapely.wkt import loads
-          >>> g = loads('POINT (0.0 0.0)')
-          >>> g.buffer(1.0).area        # 16-gon approx of a unit radius circle
-          3.1365484905459389
-          >>> g.buffer(1.0, 128).area   # 128-gon approximation
-          3.1415138011443009
-          >>> g.buffer(1.0, 3).area     # triangle approximation
-          3.0
-          >>> list(g.buffer(1.0, cap_style='square').exterior.coords)
-          [(1.0, 1.0), (1.0, -1.0), (-1.0, -1.0), (-1.0, 1.0), (1.0, 1.0)]
-          >>> g.buffer(1.0, cap_style='square').area
-          4.0
         """
         if quadsegs is not None:
             warn(
@@ -520,16 +620,27 @@ class BaseGeometry(object):
             res = quadsegs
         else:
             res = resolution
+
         if mitre_limit == 0.0:
             raise ValueError(
                 'Cannot compute offset from zero-length line segment')
+
+        if 'buffer_with_params' in self.impl:
+            params = self._lgeos.GEOSBufferParams_create()
+            try:
+                self._lgeos.GEOSBufferParams_setEndCapStyle(params, cap_style)
+                self._lgeos.GEOSBufferParams_setJoinStyle(params, join_style)
+                self._lgeos.GEOSBufferParams_setMitreLimit(params, mitre_limit)
+                self._lgeos.GEOSBufferParams_setQuadrantSegments(params, res)
+                self._lgeos.GEOSBufferParams_setSingleSided(params, single_sided)
+                return geom_factory(self.impl['buffer_with_params'](self, params, distance))
+            finally:
+                if params:
+                    self._lgeos.GEOSBufferParams_destroy(params)
+
         if cap_style == CAP_STYLE.round and join_style == JOIN_STYLE.round:
             return geom_factory(self.impl['buffer'](self, distance, res))
 
-        if 'buffer_with_style' not in self.impl:
-            raise NotImplementedError("Styled buffering not available for "
-                                      "GEOS versions < 3.2.")
-
         return geom_factory(self.impl['buffer_with_style'](self, distance, res,
                                                            cap_style,
                                                            join_style,
@@ -551,6 +662,27 @@ class BaseGeometry(object):
             op = self.impl['simplify']
         return geom_factory(op(self, tolerance))
 
+    def normalize(self):
+        """Converts geometry to normal form (or canonical form).
+
+        This method orders the coordinates, rings of a polygon and parts of
+        multi geometries consistently. Typically useful for testing purposes
+        (for example in combination with `equals_exact`).
+
+        Examples
+        --------
+        >>> from shapely.wkt import loads
+        >>> p = loads("MULTILINESTRING((0 0, 1 1), (3 3, 2 2))")
+        >>> p.normalize().wkt
+        'MULTILINESTRING ((2 2, 3 3), (0 0, 1 1))'
+        """
+        # self.impl['normalize'](self)
+        if self._geom is None:
+            raise InvalidGeometryError("Null geometry supports no operations")
+        geom_cloned = lgeos.GEOSGeom_clone(self._geom)
+        lgeos.GEOSNormalize(geom_cloned)
+        return geom_factory(geom_cloned)
+
     # Binary operations
     # -----------------
 
@@ -595,7 +727,7 @@ class BaseGeometry(object):
         """True if the geometry is closed, else False
 
         Applicable only to 1-D geometries."""
-        if self.geom_type == 'LinearRing':
+        if self.geom_type in {"LinearRing", "Polygon", "MultiPolygon"}:
             return True
         elif self.geom_type == 'LineString':
             if 'is_closed' in self.impl:
@@ -629,6 +761,10 @@ class BaseGeometry(object):
         """Returns True if the geometry covers the other, else False"""
         return bool(self.impl['covers'](self, other))
 
+    def covered_by(self, other):
+        """Returns True if the geometry is covered by the other, else False"""
+        return bool(self.impl['covered_by'](self, other))
+
     def contains(self, other):
         """Returns True if the geometry contains the other, else False"""
         return bool(self.impl['contains'](self, other))
@@ -642,7 +778,26 @@ class BaseGeometry(object):
         return bool(self.impl['disjoint'](self, other))
 
     def equals(self, other):
-        """Returns True if geometries are equal, else False"""
+        """Returns True if geometries are equal, else False.
+
+        This method considers point-set equality (or topological
+        equality), and is equivalent to (self.within(other) &
+        self.contains(other)).
+
+        Examples
+        --------
+        >>> LineString(
+        ...     [(0, 0), (2, 2)]
+        ... ).equals(
+        ...     LineString([(0, 0), (1, 1), (2, 2)])
+        ... )
+        True
+
+        Returns
+        -------
+        bool
+
+        """
         return bool(self.impl['equals'](self, other))
 
     def intersects(self, other):
@@ -662,15 +817,77 @@ class BaseGeometry(object):
         return bool(self.impl['within'](self, other))
 
     def equals_exact(self, other, tolerance):
-        """Returns True if geometries are equal to within a specified
-        tolerance"""
-        # return BinaryPredicateOp('equals_exact', self)(other, tolerance)
+        """True if geometries are equal to within a specified
+        tolerance.
+
+        Parameters
+        ----------
+        other : BaseGeometry
+            The other geometry object in this comparison.
+        tolerance : float
+            Absolute tolerance in the same units as coordinates.
+
+        This method considers coordinate equality, which requires
+        coordinates to be equal and in the same order for all components
+        of a geometry.
+
+        Because of this it is possible for "equals()" to be True for two
+        geometries and "equals_exact()" to be False.
+
+        Examples
+        --------
+        >>> LineString(
+        ...     [(0, 0), (2, 2)]
+        ... ).equals_exact(
+        ...     LineString([(0, 0), (1, 1), (2, 2)]),
+        ...     1e-6
+        ... )
+        False
+
+        Returns
+        -------
+        bool
+
+        """
         return bool(self.impl['equals_exact'](self, other, tolerance))
 
     def almost_equals(self, other, decimal=6):
-        """Returns True if geometries are equal at all coordinates to a
-        specified decimal place"""
-        return self.equals_exact(other, 0.5 * 10**(-decimal))
+        """True if geometries are equal at all coordinates to a
+        specified decimal place.
+
+        .. deprecated:: 1.8.0 The 'almost_equals()' method is deprecated
+            and will be removed in Shapely 2.0 because the name is
+            confusing. The 'equals_exact()' method should be used
+            instead.
+
+        Refers to approximate coordinate equality, which requires
+        coordinates to be approximately equal and in the same order for
+        all components of a geometry.
+
+        Because of this it is possible for "equals()" to be True for two
+        geometries and "almost_equals()" to be False.
+
+        Examples
+        --------
+        >>> LineString(
+        ...     [(0, 0), (2, 2)]
+        ... ).equals_exact(
+        ...     LineString([(0, 0), (1, 1), (2, 2)]),
+        ...     1e-6
+        ... )
+        False
+
+        Returns
+        -------
+        bool
+
+        """
+        warn(
+            "The 'almost_equals()' method is deprecated and will be removed in Shapely 2.0",
+            ShapelyDeprecationWarning,
+            stacklevel=2,
+        )
+        return self.equals_exact(other, 0.5 * 10 ** (-decimal))
 
     def relate_pattern(self, other, pattern):
         """Returns True if the DE-9IM string code for the relationship between
@@ -696,9 +913,13 @@ class BaseGeometry(object):
         return op(self, other)
 
     @delegated
+    @exceptNull
     def interpolate(self, distance, normalized=False):
         """Return a point at the specified distance along a linear geometry
 
+        Negative length values are taken as measured in the reverse
+        direction from the end of the geometry. Out-of-range index
+        values are handled by clamping them to the valid range of values.
         If the normalized arg is True, the distance will be interpreted as a
         fraction of the geometry's length.
         """
@@ -723,8 +944,8 @@ class BaseMultipartGeometry(BaseGeometry):
     @property
     def __array_interface__(self):
         """Provide the Numpy array protocol."""
-        raise NotImplementedError("Multi-part geometries do not themselves "
-                                  "provide the array interface")
+        raise AttributeError("Multi-part geometries do not themselves "
+                             "provide the array interface")
 
     def _get_coords(self):
         raise NotImplementedError("Sub-geometries may have coordinate "
@@ -745,19 +966,47 @@ class BaseMultipartGeometry(BaseGeometry):
             return []
         return GeometrySequence(self, self.shape_factory)
 
+    def __bool__(self):
+        return self.is_empty is False
+
     def __iter__(self):
+        """
+        .. deprecated:: 1.8
+           Iteration over multi-part geometries is deprecated and will be removed in
+           Shapely 2.0. Use the `geoms` property to access the constituent parts of
+           a multi-part geometry.
+        """
+        warn(
+            "Iteration over multi-part geometries is deprecated and will be removed in "
+            "Shapely 2.0. Use the `geoms` property to access the constituent parts of "
+            "a multi-part geometry.", ShapelyDeprecationWarning, stacklevel=2)
         if not self.is_empty:
             return iter(self.geoms)
         else:
             return iter([])
 
     def __len__(self):
+        warn(
+            "__len__ for multi-part geometries is deprecated and will be removed in "
+            "Shapely 2.0. Check the length of the `geoms` property instead to get the "
+            " number of parts of a multi-part geometry.",
+            ShapelyDeprecationWarning, stacklevel=2)
         if not self.is_empty:
             return len(self.geoms)
         else:
             return 0
 
     def __getitem__(self, index):
+        """
+        .. deprecated:: 1.8
+           __getitem__ for multi-part geometries is deprecated and will be removed in
+           Shapely 2.0. Use the `geoms` property to access the constituent parts of
+           a multi-part geometry.
+        """
+        warn(
+            "__getitem__ for multi-part geometries is deprecated and will be removed in "
+            "Shapely 2.0. Use the `geoms` property to access the constituent parts of "
+            "a multi-part geometry.", ShapelyDeprecationWarning, stacklevel=2)
         if not self.is_empty:
             return self.geoms[index]
         else:
@@ -765,15 +1014,15 @@ class BaseMultipartGeometry(BaseGeometry):
 
     def __eq__(self, other):
         return (
-            isinstance(other, self.__class__) and
-            len(self) == len(other) and
-            all(x == y for x, y in zip(self, other))
+            type(other) == type(self) and
+            len(self.geoms) == len(other.geoms) and
+            all(x == y for x, y in zip(self.geoms, other.geoms))
         )
 
     def __ne__(self, other):
         return not self.__eq__(other)
 
-    __hash__ = object.__hash__
+    __hash__ = None
 
     def svg(self, scale_factor=1., color=None):
         """Returns a group of SVG elements for the multipart geometry.
@@ -791,11 +1040,11 @@ class BaseMultipartGeometry(BaseGeometry):
         if color is None:
             color = "#66cc99" if self.is_valid else "#ff3333"
         return '<g>' + \
-            ''.join(p.svg(scale_factor, color) for p in self) + \
+            ''.join(p.svg(scale_factor, color) for p in self.geoms) + \
             '</g>'
 
 
-class GeometrySequence(object):
+class GeometrySequence:
     """
     Iterative access to members of a homogeneous multipart geometry.
     """
@@ -826,7 +1075,7 @@ class GeometrySequence(object):
     def _get_geom_item(self, i):
         g = self.shape_factory()
         g._other_owned = True
-        g._geom = lgeos.GEOSGetGeometryN(self._geom, i)
+        g._set_geom(lgeos.GEOSGetGeometryN(self._geom, i))
         g._ndim = self._ndim
         g.__p__ = self
         return g
@@ -853,8 +1102,8 @@ class GeometrySequence(object):
             return self._get_geom_item(i)
         elif isinstance(key, slice):
             if type(self) == HeterogeneousGeometrySequence:
-                raise TypeError(
-                    "Heterogenous geometry collections are not sliceable")
+                raise GeometryTypeError(
+                    "Heterogeneous geometry collections are not sliceable")
             res = []
             start, stop, stride = key.indices(m)
             for i in range(start, stop, stride):
@@ -878,7 +1127,7 @@ class HeterogeneousGeometrySequence(GeometrySequence):
     """
 
     def __init__(self, parent):
-        super(HeterogeneousGeometrySequence, self).__init__(parent, None)
+        super().__init__(parent, None)
 
     def _get_geom_item(self, i):
         sub = lgeos.GEOSGetGeometryN(self._geom, i)
@@ -887,10 +1136,7 @@ class HeterogeneousGeometrySequence(GeometrySequence):
         return g
 
 
-def _test():
-    """Test runner"""
-    import doctest
-    doctest.testmod()
-
-if __name__ == "__main__":
-    _test()
+class EmptyGeometry(BaseGeometry):
+    def __init__(self):
+        """Create an empty geometry."""
+        BaseGeometry.__init__(self)
