@@ -1,13 +1,11 @@
 """Collections of linestrings and related utilities
 """
 
-import sys
-
-if sys.version_info[0] < 3:
-    range = xrange
-
-from ctypes import c_double, c_void_p, cast, POINTER
+from ctypes import c_void_p, cast
+import warnings
 
+from shapely.ctypes_declarations import c_geom_p
+from shapely.errors import EmptyPartError, ShapelyDeprecationWarning
 from shapely.geos import lgeos
 from shapely.geometry.base import BaseMultipartGeometry, geos_geom_from_py
 from shapely.geometry import linestring
@@ -43,13 +41,15 @@ class MultiLineString(BaseMultipartGeometry):
 
           >>> lines = MultiLineString( [[[0.0, 0.0], [1.0, 2.0]]] )
         """
-        super(MultiLineString, self).__init__()
+        super().__init__()
 
         if not lines:
             # allow creation of empty multilinestrings, to support unpickling
             pass
         else:
-            self._geom, self._ndim = geos_multilinestring_from_py(lines)
+            geom, n = geos_multilinestring_from_py(lines)
+            self._set_geom(geom)
+            self._ndim = n
 
     def shape_factory(self, *args):
         return linestring.LineString(*args)
@@ -61,26 +61,26 @@ class MultiLineString(BaseMultipartGeometry):
             'coordinates': tuple(tuple(c for c in g.coords) for g in self.geoms)
             }
 
-    def svg(self, scale_factor=1.):
+    def svg(self, scale_factor=1., stroke_color=None, opacity=None):
+        """Returns a group of SVG polyline elements for the LineString geometry.
+
+        Parameters
+        ==========
+        scale_factor : float
+            Multiplication factor for the SVG stroke-width.  Default is 1.
+        stroke_color : str, optional
+            Hex string for stroke color. Default is to use "#66cc99" if
+            geometry is valid, and "#ff3333" if invalid.
+        opacity : float
+            Float number between 0 and 1 for color opacity. Default value is 0.8
         """
-        SVG representation of the geometry. Scale factor is multiplied by
-        the size of the SVG symbol so it can be scaled consistently for a
-        consistent appearance based on the canvas size.
-        """
-        parts = []
-        for part in self.geoms:
-            pnt_format = " ".join(["{0},{1}".format(*c) for c in part.coords])
-            parts.append("""<polyline
-                fill="none"
-                stroke="{2}"
-                stroke-width={1}
-                points="{0}"
-                opacity=".8"
-                />""".format(
-                    pnt_format,
-                    2.*scale_factor,
-                    "#66cc99" if self.is_valid else "#ff3333"))
-        return "\n".join(parts)
+        if self.is_empty:
+            return '<g />'
+        if stroke_color is None:
+            stroke_color = "#66cc99" if self.is_valid else "#ff3333"
+        return '<g>' + \
+            ''.join(p.svg(scale_factor, stroke_color, opacity) for p in self.geoms) + \
+            '</g>'
 
 
 class MultiLineStringAdapter(CachingGeometryProxy, MultiLineString):
@@ -89,6 +89,12 @@ class MultiLineStringAdapter(CachingGeometryProxy, MultiLineString):
     _other_owned = False
 
     def __init__(self, context):
+        warnings.warn(
+            "The proxy geometries (through the 'asShape()', 'asMultiLineString()' "
+            "or 'MultiLineStringAdapter()' constructors) are deprecated and will be "
+            "removed in Shapely 2.0. Use the 'shape()' function or the "
+            "standard 'MultiLineString()' constructor instead.",
+            ShapelyDeprecationWarning, stacklevel=4)
         self.context = context
         self.factory = geos_multilinestring_from_py
 
@@ -111,9 +117,9 @@ def asMultiLineString(context):
 
 
 def geos_multilinestring_from_py(ob):
-    # ob must be either a MultiLineString, a sequence, or 
+    # ob must be either a MultiLineString, a sequence, or
     # array of sequences or arrays
-    
+
     if isinstance(ob, MultiLineString):
          return geos_geom_from_py(ob)
 
@@ -129,20 +135,15 @@ def geos_multilinestring_from_py(ob):
         raise ValueError("Invalid coordinate dimensionality")
 
     # Array of pointers to point geometries
-    subs = (c_void_p * L)()
-    
+    subs = (c_geom_p * L)()
+
     # add to coordinate sequence
     for l in range(L):
         geom, ndims = linestring.geos_linestring_from_py(obs[l])
-        subs[l] = cast(geom, c_void_p)
-            
+
+        if lgeos.GEOSisEmpty(geom):
+            raise EmptyPartError("Can't create MultiLineString with empty component")
+
+        subs[l] = cast(geom, c_geom_p)
+
     return (lgeos.GEOSGeom_createCollection(5, subs, L), N)
-
-
-# Test runner
-def _test():
-    import doctest
-    doctest.testmod()
-
-if __name__ == "__main__":
-    _test()
