@@ -1,19 +1,14 @@
 """Collections of points and related utilities
 """
 
-import sys
+from ctypes import byref, c_double, c_void_p, cast
+import warnings
 
-if sys.version_info[0] < 3:
-    range = xrange
-
-from ctypes import byref, c_double, c_void_p, cast, POINTER
-from ctypes import ArgumentError
-
-from shapely.coords import required
+from shapely.ctypes_declarations import c_geom_p
+from shapely.errors import EmptyPartError, ShapelyDeprecationWarning
 from shapely.geos import lgeos
 from shapely.geometry.base import (
-    BaseMultipartGeometry, exceptNull, geos_geom_from_py
-)
+    BaseMultipartGeometry, exceptNull, geos_geom_from_py)
 from shapely.geometry import point
 from shapely.geometry.proxy import CachingGeometryProxy
 
@@ -39,25 +34,28 @@ class MultiPoint(BaseMultipartGeometry):
         points : sequence
             A sequence of (x, y [,z]) numeric coordinate pairs or triples or a
             sequence of objects that implement the numpy array interface,
-            including instaces of Point.
+            including instances of Point.
 
         Example
         -------
         Construct a 2 point collection
 
+          >>> from shapely.geometry import Point
           >>> ob = MultiPoint([[0.0, 0.0], [1.0, 2.0]])
           >>> len(ob.geoms)
           2
           >>> type(ob.geoms[0]) == Point
           True
         """
-        super(MultiPoint, self).__init__()
+        super().__init__()
 
-        if points is None:
+        if points is None or (not isinstance(points, MultiPoint) and len(points) == 0):
             # allow creation of empty multipoints, to support unpickling
             pass
         else:
-            self._geom, self._ndim = geos_multipoint_from_py(points)
+            geom, n = geos_multipoint_from_py(points)
+            self._set_geom(geom)
+            self._ndim = n
 
     def shape_factory(self, *args):
         return point.Point(*args)
@@ -69,7 +67,7 @@ class MultiPoint(BaseMultipartGeometry):
             'coordinates': tuple([g.coords[0] for g in self.geoms])
             }
 
-    def svg(self, scale_factor=1., fill_color=None):
+    def svg(self, scale_factor=1., fill_color=None, opacity=None):
         """Returns a group of SVG circle elements for the MultiPoint geometry.
 
         Parameters
@@ -79,18 +77,20 @@ class MultiPoint(BaseMultipartGeometry):
         fill_color : str, optional
             Hex string for fill color. Default is to use "#66cc99" if
             geometry is valid, and "#ff3333" if invalid.
+        opacity : float
+            Float number between 0 and 1 for color opacity. Default value is 0.6
         """
         if self.is_empty:
             return '<g />'
         if fill_color is None:
             fill_color = "#66cc99" if self.is_valid else "#ff3333"
         return '<g>' + \
-            ''.join(p.svg(scale_factor, fill_color) for p in self) + \
+            ''.join(p.svg(scale_factor, fill_color, opacity) for p in self.geoms) + \
             '</g>'
 
     @property
     @exceptNull
-    def ctypes(self):
+    def _ctypes(self):
         if not self._ctypes_data:
             temp = c_double()
             n = self._ndim
@@ -98,7 +98,7 @@ class MultiPoint(BaseMultipartGeometry):
             array_type = c_double * (m * n)
             data = array_type()
             for i in range(m):
-                g = self.geoms[i]._geom    
+                g = self.geoms[i]._geom
                 cs = lgeos.GEOSGeom_getCoordSeq(g)
                 lgeos.GEOSCoordSeq_getX(cs, 0, byref(temp))
                 data[n*i] = temp.value
@@ -110,13 +110,36 @@ class MultiPoint(BaseMultipartGeometry):
             self._ctypes_data = data
         return self._ctypes_data
 
+    @property
+    def ctypes(self):
+        warnings.warn(
+            "Accessing the 'ctypes' attribute is deprecated,"
+            " and will not be possible any more in Shapely 2.0",
+            ShapelyDeprecationWarning, stacklevel=2)
+        return self._ctypes
+
     @exceptNull
-    def array_interface(self):
+    def _array_interface(self):
         """Provide the Numpy array protocol."""
-        ai = self.array_interface_base
+        ai = self._array_interface_base
         ai.update({'shape': (len(self.geoms), self._ndim)})
         return ai
-    __array_interface__ = property(array_interface)
+
+    def array_interface(self):
+        """Provide the Numpy array protocol."""
+        warnings.warn(
+            "The 'array_interface' method is deprecated and will be removed "
+            "in Shapely 2.0.",
+            ShapelyDeprecationWarning, stacklevel=2)
+        return self._array_interface()
+
+    @property
+    def __array_interface__(self):
+        warnings.warn(
+            "The array interface is deprecated and will no longer work in "
+            "Shapely 2.0. Convert the '.coords' to a numpy array instead.",
+            ShapelyDeprecationWarning, stacklevel=3)
+        return self._array_interface()
 
 
 class MultiPointAdapter(CachingGeometryProxy, MultiPoint):
@@ -125,6 +148,12 @@ class MultiPointAdapter(CachingGeometryProxy, MultiPoint):
     _other_owned = False
 
     def __init__(self, context):
+        warnings.warn(
+            "The proxy geometries (through the 'asShape()', 'asMultiPoint()' "
+            "or 'MultiPointAdapter()' constructors) are deprecated and will be "
+            "removed in Shapely 2.0. Use the 'shape()' function or the "
+            "standard 'MultiPoint()' constructor instead.",
+            ShapelyDeprecationWarning, stacklevel=4)
         self.context = context
         self.factory = geos_multipoint_from_py
 
@@ -158,57 +187,24 @@ def geos_multipoint_from_py(ob):
     if isinstance(ob, MultiPoint):
         return geos_geom_from_py(ob)
 
-    # If numpy is present, we use numpy.require to ensure that we have a
-    # C-continguous array that owns its data. View data will be copied.
-    ob = required(ob)
+    m = len(ob)
     try:
-        # From array protocol
-        array = ob.__array_interface__
-        assert len(array['shape']) == 2
-        m = array['shape'][0]
-        n = array['shape'][1]
-        assert m >= 1
-        assert n == 2 or n == 3
+        n = len(ob[0])
+    except TypeError:
+        n = ob[0]._ndim
+    assert n == 2 or n == 3
 
-        # Make pointer to the coordinate array
-        if isinstance(array['data'], tuple):
-            # numpy tuple (addr, read-only)
-            cp = cast(array['data'][0], POINTER(c_double))
-        else:
-            cp = array['data']
+    # Array of pointers to point geometries
+    subs = (c_geom_p * m)()
 
-        # Array of pointers to sub-geometries
-        subs = (c_void_p * m)()
+    # add to coordinate sequence
+    for i in range(m):
+        coords = ob[i]
+        geom, ndims = point.geos_point_from_py(coords)
 
-        for i in range(m):
-            geom, ndims = point.geos_point_from_py(cp[n*i:n*i+2])
-            subs[i] = cast(geom, c_void_p)
+        if lgeos.GEOSisEmpty(geom):
+            raise EmptyPartError("Can't create MultiPoint with empty component")
 
-    except AttributeError:
-        # Fall back on list
-        m = len(ob)
-        try:
-            n = len(ob[0])
-        except TypeError:
-            n = ob[0]._ndim
-        assert n == 2 or n == 3
+        subs[i] = cast(geom, c_geom_p)
 
-        # Array of pointers to point geometries
-        subs = (c_void_p * m)()
-        
-        # add to coordinate sequence
-        for i in range(m):
-            coords = ob[i]
-            geom, ndims = point.geos_point_from_py(coords)
-            subs[i] = cast(geom, c_void_p)
-            
     return lgeos.GEOSGeom_createCollection(4, subs, m), n
-
-# Test runner
-def _test():
-    import doctest
-    doctest.testmod()
-
-
-if __name__ == "__main__":
-    _test()
