@@ -1,13 +1,11 @@
 """Collections of polygons and related utilities
 """
 
-import sys
-
-if sys.version_info[0] < 3:
-    range = xrange
-
 from ctypes import c_void_p, cast
+import warnings
 
+from shapely.ctypes_declarations import c_geom_p
+from shapely.errors import ShapelyDeprecationWarning
 from shapely.geos import lgeos
 from shapely.geometry.base import BaseMultipartGeometry, geos_geom_from_py
 from shapely.geometry import polygon
@@ -19,10 +17,10 @@ __all__ = ['MultiPolygon', 'asMultiPolygon']
 class MultiPolygon(BaseMultipartGeometry):
 
     """A collection of one or more polygons
-    
+
     If component polygons overlap the collection is `invalid` and some
     operations on it may fail.
-    
+
     Attributes
     ----------
     geoms : sequence
@@ -42,9 +40,10 @@ class MultiPolygon(BaseMultipartGeometry):
         -------
         Construct a collection from a sequence of coordinate tuples
 
+          >>> from shapely.geometry import Polygon
           >>> ob = MultiPolygon( [
           ...     (
-          ...     ((0.0, 0.0), (0.0, 1.0), (1.0, 1.0), (1.0, 0.0)), 
+          ...     ((0.0, 0.0), (0.0, 1.0), (1.0, 1.0), (1.0, 0.0)),
           ...     [((0.1,0.1), (0.1,0.2), (0.2,0.2), (0.2,0.1))]
           ...     )
           ... ] )
@@ -53,15 +52,17 @@ class MultiPolygon(BaseMultipartGeometry):
           >>> type(ob.geoms[0]) == Polygon
           True
         """
-        super(MultiPolygon, self).__init__()
+        super().__init__()
 
         if not polygons:
             # allow creation of empty multipolygons, to support unpickling
-            pass
+            return
         elif context_type == 'polygons':
-            self._geom, self._ndim = geos_multipolygon_from_polygons(polygons)
+            geom, n = geos_multipolygon_from_polygons(polygons)
         elif context_type == 'geojson':
-            self._geom, self._ndim = geos_multipolygon_from_py(polygons)
+            geom, n = geos_multipolygon_from_py(polygons)
+        self._set_geom(geom)
+        self._ndim = n
 
     def shape_factory(self, *args):
         return polygon.Polygon(*args)
@@ -80,7 +81,7 @@ class MultiPolygon(BaseMultipartGeometry):
             'coordinates': allcoords
             }
 
-    def svg(self, scale_factor=1., fill_color=None):
+    def svg(self, scale_factor=1., fill_color=None, opacity=None):
         """Returns group of SVG path elements for the MultiPolygon geometry.
 
         Parameters
@@ -90,22 +91,30 @@ class MultiPolygon(BaseMultipartGeometry):
         fill_color : str, optional
             Hex string for fill color. Default is to use "#66cc99" if
             geometry is valid, and "#ff3333" if invalid.
+        opacity : float
+            Float number between 0 and 1 for color opacity. Default value is 0.6
         """
         if self.is_empty:
             return '<g />'
         if fill_color is None:
             fill_color = "#66cc99" if self.is_valid else "#ff3333"
         return '<g>' + \
-            ''.join(p.svg(scale_factor, fill_color) for p in self) + \
+            ''.join(p.svg(scale_factor, fill_color, opacity) for p in self.geoms) + \
             '</g>'
 
 
 class MultiPolygonAdapter(CachingGeometryProxy, MultiPolygon):
-    
+
     context = None
     _other_owned = False
 
     def __init__(self, context, context_type='polygons'):
+        warnings.warn(
+            "The proxy geometries (through the 'asShape()', 'asMultiPolygon()' "
+            "or 'MultiPolygonAdapter()' constructors) are deprecated and will be "
+            "removed in Shapely 2.0. Use the 'shape()' function or the "
+            "standard 'MultiPolygon()' constructor instead.",
+            ShapelyDeprecationWarning, stacklevel=3)
         self.context = context
         if context_type == 'geojson':
             self.factory = geos_multipolygon_from_py
@@ -134,56 +143,67 @@ def geos_multipolygon_from_py(ob):
     """ob must provide Python geo interface coordinates."""
     L = len(ob)
     assert L >= 1
-    
+
     N = len(ob[0][0][0])
     assert N == 2 or N == 3
 
-    subs = (c_void_p * L)()
+    subs = (c_geom_p * L)()
     for l in range(L):
         geom, ndims = polygon.geos_polygon_from_py(ob[l][0], ob[l][1:])
-        subs[l] = cast(geom, c_void_p)
-            
+        subs[l] = cast(geom, c_geom_p)
+
     return (lgeos.GEOSGeom_createCollection(6, subs, L), N)
 
 
-def geos_multipolygon_from_polygons(ob):
-    """
-    ob must be either a MultiPolygon, sequence or array of sequences 
-    or arrays.
-    
-    """
-    if isinstance(ob, MultiPolygon):
-        return geos_geom_from_py(ob)
+def geos_multipolygon_from_polygons(arg):
+    """Creates a GEOS multipolygon from a sequence of polygon-like objects.
 
-    obs = getattr(ob, 'geoms', None) or ob
+    Parameters
+    ----------
+    arg : sequence or MultiPolygon
+
+    Returns
+    -------
+    int
+        Pointer to a GEOS multipolygon.
+
+    """
+    if isinstance(arg, MultiPolygon):
+        return geos_geom_from_py(arg)
+
+    obs = getattr(arg, 'geoms', arg)
+    obs = [ob for ob in obs
+           if ob and not (isinstance(ob, polygon.Polygon) and ob.is_empty)]
     L = len(obs)
-    assert L >= 1
-    
+
+    # Bail immediately if we have no input points.
+    if L <= 0:
+        return (lgeos.GEOSGeom_createEmptyCollection(6), 3)
+
+    # This function does not accept sequences of MultiPolygons: there is
+    # no implicit flattening.
+    if isinstance(obs[0], MultiPolygon):
+        raise ValueError("Sequences of multi-polygons are not valid arguments")
+
     exemplar = obs[0]
     try:
         N = len(exemplar[0][0])
     except TypeError:
         N = exemplar._ndim
-    
+
     assert N == 2 or N == 3
 
-    subs = (c_void_p * L)()
-    for l in range(L):
-        shell = getattr(obs[l], 'exterior', None)
-        if shell is None:
-            shell = obs[l][0]
-        holes = getattr(obs[l], 'interiors', None)
-        if holes is None:
-            holes =  obs[l][1]
+    subs = (c_geom_p * L)()
+
+    for i, ob in enumerate(obs):
+        if isinstance(ob, polygon.Polygon):
+            shell = ob.exterior
+            holes = ob.interiors
+        else:
+            shell = ob[0]
+            holes = ob[1]
+
         geom, ndims = polygon.geos_polygon_from_py(shell, holes)
-        subs[l] = cast(geom, c_void_p)
-            
+        subs[i] = cast(geom, c_geom_p)
+
     return (lgeos.GEOSGeom_createCollection(6, subs, L), N)
-
-# Test runner
-def _test():
-    import doctest
-    doctest.testmod()
-
-if __name__ == "__main__":
-    _test()
